/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "cmc_typ.h"
#include "blosum62.h"

BooLean	cmc_typ::SampleHpt0(FILE *fp,Int4 SampledCol, Int4 Root, wdg_typ &Tree)
// WARNING: this routine assumes direct correspondence between column and rows!!
// So that the diagonal equates rows (sets) with columns (FD-categories)!!
{
    // 3. Change that cell and recompute LPR.
    // 4. If this improves the LPR then accept the change, else revert back... 

    Int4	g,n,r,c,x;
    char	state;
    double	olpr,nlpr;
    Int4	row_neg,row_pos,row_omit,col_neg,col_pos,col_omit;
    Int4	sq,num_sq = NumSeqsCMSA(MainCMA);

print_error("WARNING: SampleHpt0( ) needs to be checked!!!!!!!!!");
    wdg_typ NewTree=0;	// NewTree to merge with old tree...
    if(Tree){ NewTree=MkWdgraph(WdgraphN(Tree),WdgraphM(Tree)); }

    assert(Hpt->NumSets() == Hpt->NumBPPS() +1); // assumes direct correspondence between column and rows!!
    // Start at n=2 because don't want to modify main set...
    for(n=2; n<= Hpt->NumBPPS(); n++){	// 2. Find a column to work on.

//************ Testing this for cdhBPPS routine: adding an internal node....
	if(SampledCol && SampledCol != n) continue;

	g=n; 	// assume direct correspondence between column and rows!!
	if(Hpt->TypeOfSet(g) != '?') continue; // find a misc. column 
    	// 1. Found miscellaneous column to change within the hpt.
	char    **HP=HyperPartition;
        for(g=1; g < Hpt->NumSets(); g++){		// for each row... Don't look at Rejects...
	    if(Hpt->Cell(g,n) == '+') continue;	// don't look at positive cells right now...
	    if(Hpt->Cell(g,n) == 'o') continue;	// don't look at omitted cells right now...
	    assert(Hpt->Cell(g,n) == '-');
	    Hpt->RowsInColumn(n,row_neg,row_pos,row_omit);   // count # '-' cells in column n.
	    if(row_neg < 2) continue; 
	    olpr = CalcTotalLPR( );
	    for(sq=1; sq <= num_sq; sq++){
		    if(MemberSet(sq,GrpSet[g])){ che[n]->SetPartition('+',sq); }
	    } nlpr=CalcTotalLPR( );
	    double d=nlpr-olpr;
#if 0
	    if(d < 10 && d > -10.0){ HP[g][n] = 'o'; // etc...  
	    }
#endif
	    if(d < 10.0){	// then revert back to previous..
	      if(fp) fprintf(fp,"Failed to change HP[%d][%d] from '-' to '+' (delta=%.3f).\n",g,n,d);
		for(sq=1; sq <= num_sq; sq++){
		     if(MemberSet(sq,GrpSet[g])){ che[n]->SetPartition('-',sq); } }
	    } else {	// accept change...
	      BooLean OkayToAdd=TRUE;
	      if(Tree){		// Then add an edge from set in row g to the misc node n within the tree.
		// WARNING: This assumes that Hpt was first generated from a tree! Won't work otherwise.
		x=(Int4) floor(d+0.5);
		if(Hpt->AddEdgeToTree(n,g,Root,x,Tree,NewTree) == 0) OkayToAdd=FALSE;
		// PrintNewickTreeWDG(stderr,Root, Tree); // need to pass in Root
	      } // Add edges to the input Tree.
	      if(OkayToAdd){ 
	        if(fp) fprintf(fp,"Changed HP[%d][%d] from '-' to '+' (delta=%.3f).\n",g,n,d);
	        Hpt->Change('-','+',g,n); ReSetRelations(); assert(HP[g][n] == '+');
	      } else { // revert back...
		for(sq=1; sq <= num_sq; sq++){ if(MemberSet(sq,GrpSet[g])){ che[n]->SetPartition('-',sq); } }
	      }

	      // Change the omitted columns for this set which is now included in misc set...
	      // 1. Find the rows to be omitted for this set in column g == row n set only.
	      for(olpr=nlpr,r=1; r < Hpt->NumSets(); r++){	// skip Rejected set..
		if(r == g) continue;
		if(HP[r][n] == '+'){	// if set 'r' also is in the FG for column n...
		    if(HP[g][r] == 'o'){   // & if set g is omitted from the BG for set r.
		      // then see whether this should be included in the BG.
		      for(sq=1; sq <= num_sq; sq++){
			if(MemberSet(sq,GrpSet[g])){ che[r]->SetPartition('-',sq); }
		      } nlpr=CalcTotalLPR( ); d=nlpr-olpr;
		      if(d < 10.0){  // Then revert..
	      		if(fp) fprintf(fp,"Failed to change HP[%d][%d] from 'o' to '-' (delta=%.3f)\n",g,r,d);
			for(sq=1; sq <= num_sq; sq++){
			   if(MemberSet(sq,GrpSet[g])){ che[r]->SetPartition('o',sq); } }
		      } else {
	      		if(fp) fprintf(fp,"Changed HP[%d][%d] from 'o' to '-' (delta=%.3f)\n",g,r,d);
	      		Hpt->Change('o','-',g,r); ReSetRelations(); assert(HP[g][r] == '-');
			olpr=nlpr;
		      }
		    }
		}
	      }
	      // Change '-' rows to 'o' in column g (== row g) if those rows == '-' for column n.
	      // 1. Find the rows to be changed to 'o' in column g.
	      for(r=1; r < Hpt->NumSets(); r++){	// skip Rejected set..
		if(r == g) continue;
		if(HP[r][n] == '-' && HP[r][g] == '-'){	
		      for(sq=1; sq <= num_sq; sq++){
			if(MemberSet(sq,GrpSet[r])){ che[g]->SetPartition('-',sq); }
		      } nlpr=CalcTotalLPR( ); d=nlpr-olpr;
		      if(FALSE && d < 0.0){  // Then revert..
	      		if(fp) fprintf(fp,"Failed to change HP[%d][%d] from '-' to 'o' (delta=%.3f)\n",g,r,d);
			for(sq=1; sq <= num_sq; sq++){
			   if(MemberSet(sq,GrpSet[r])){ che[g]->SetPartition('o',sq); } }
		      } else {
	      		if(fp) fprintf(fp,"Changed HP[%d][%d] from '-' to 'o' (delta=%.3f)\n",g,r,d);
	      		Hpt->Change('-','o',r,g); ReSetRelations(); assert(HP[r][g] == 'o');
			olpr=nlpr;
		      }
		}
	      }
	      // Change 'o' rows to '-' in column n (== row g) if those rows == 'o' for column g.
	      for(r=1; r < Hpt->NumSets(); r++){	// skip Rejected set..
		if(r == g) continue;
		if(HP[r][n] == '+' && HP[g][r] == 'o'){	
		      for(sq=1; sq <= num_sq; sq++){
			if(MemberSet(sq,GrpSet[g])){ che[r]->SetPartition('-',sq); }
		      } nlpr=CalcTotalLPR( ); d=nlpr-olpr;
		      if(FALSE && d < 0.0){  // Then revert..
	      		if(fp) fprintf(fp,"Failed to change HP[%d][%d] from 'o' to '-' (delta=%.3f)\n",g,r,d);
			for(sq=1; sq <= num_sq; sq++){
			   if(MemberSet(sq,GrpSet[g])){ che[r]->SetPartition('o',sq); } }
		      } else {
	      		if(fp) fprintf(fp,"Changed HP[%d][%d] from 'o' to '-' (delta=%.3f)\n",g,r,d);
	      		Hpt->Change('o','-',g,r); ReSetRelations(); assert(HP[g][r] == '-');
			olpr=nlpr;
		      }
		}
	      }
	    }
	}
    } 
    if(Tree){ Hpt->MergeOldTreeIntoNew(Root,Tree,NewTree); NilWdgraph(Tree); Tree=NewTree; }
    CalcTotalLPR( );
}

