/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "bTreeNode.h"

bTreeNode::bTreeNode() {} /* Default Constructor should not be used, one must set the id num */

bTreeNode::bTreeNode( Int4 num )
  : id(num), leftDis(0.), rightDis(0.), leftPtr(0), rightPtr(0), parentPtr(0), numLeaves(1) {}

bTreeNode::bTreeNode( bTreeNode *left, bTreeNode *right, double leftD, double rightD, Int4 num ) 
  : id(num), leftDis(leftD), rightDis(rightD), leftPtr(left), rightPtr(right), parentPtr(0)
{
  assert ( left != 0 || right != 0 );
  leftPtr->parentPtr = this;
  rightPtr->parentPtr = this;
  numLeaves = leftPtr->numLeaves + rightPtr->numLeaves;
}
bTreeNode::~bTreeNode() 
{
  NilbTreeHelper( &leftPtr );
  NilbTreeHelper( &rightPtr );
}
void bTreeNode::inOrderTraversal() 
{
  inOrderHelper(this);
}
void bTreeNode::getTreeFile( FILE *file )
{
  if ( file == 0 ) {
    return;
  }
  char holder;
  double tempDis;
  bTreeNode *tempPtr, *temp2Ptr;

  NilbTreeHelper(&leftPtr);
  NilbTreeHelper(&rightPtr);
  leftDis = rightDis = id = 0;
  do {
    holder = fgetc(file);
  } while (holder == ' ' || holder == '\n' || holder == '\t' || holder == '\r' );
  if ( holder == '(' ) {
    leftDis = getTreeFileHelper ( &file, &leftPtr ); 
    rightDis = getTreeFileHelper( &file, &rightPtr );
    fseek( file, -1L, SEEK_CUR );
    holder = fgetc(file);
    if ( holder == ',' ) {
      tempDis = getTreeFileHelper( &file, &tempPtr );
      temp2Ptr = new bTreeNode( rightPtr, tempPtr, rightDis, tempDis, -1 );
      temp2Ptr->numLeaves = rightPtr->numLeaves + tempPtr->numLeaves;
      rightPtr = temp2Ptr;
      rightDis = 0;
    }
    leftPtr->parentPtr = this;
    rightPtr->parentPtr = this;
    numLeaves = rightPtr->numLeaves + leftPtr->numLeaves;
    holder = fgetc(file);
    if ( holder == ';' ) {
      printf("getTreeFile: The tree has been inputted.\n");
    } else {
      printf("Error.\n");
    }
  } else {
    printf("Error.\n");
  }
}
double bTreeNode::getTreeFileHelper( FILE **file, bTreeNode **bnode )
{
  double doubleHolder;
  char holder = fgetc((*file));
  (*bnode) = new bTreeNode();
  if ( holder == '(' ) {
    (*bnode)->leftDis = getTreeFileHelper ( file, &((*bnode)->leftPtr) );
    (*bnode)->leftPtr->parentPtr = (*bnode);
    (*bnode)->rightDis = getTreeFileHelper ( file, &((*bnode)->rightPtr) );
    (*bnode)->rightPtr->parentPtr = (*bnode);
    (*bnode)->numLeaves = (*bnode)->leftPtr->numLeaves + (*bnode)->rightPtr->numLeaves;
  } else {
    fseek(*file, -1L, SEEK_CUR);
    fscanf((*file), "%d", &((*bnode)->id));
    (*bnode)->numLeaves = 1;
  }
  fscanf((*file), ":%lf", &doubleHolder );
  fseek(*file, 1L, SEEK_CUR);
  return(doubleHolder);
}
void bTreeNode::putTreeFile( FILE *file )
{
  if ( file == 0 || leftPtr == 0 || rightPtr == 0 ) {
    return;
  }
  fprintf(file, "(");
  putTreeFileHelper( &file, leftPtr, leftDis );
  fprintf(file, ",");
  if ( rightDis == 0. ) {
    putTreeFileHelper( &file, rightPtr->leftPtr, rightPtr->leftDis );
    fprintf(file, ",");
    putTreeFileHelper( &file, rightPtr->rightPtr, rightPtr->rightDis );
  } else {
    putTreeFileHelper ( &file, rightPtr, rightDis );
  }
  fprintf(file, ");");
  printf("putTreeFile: The file has been created.\n");
}
void bTreeNode::putTreeFileHelper ( FILE ** file, bTreeNode *bnode, double dis )
{
  if ( bnode->leftPtr == 0 && bnode->rightPtr == 0 ) {
    fprintf((*file),"%d", bnode->id);
  } else {
    fprintf((*file), "(");
    putTreeFileHelper( file, bnode->leftPtr, bnode->leftDis );
    fprintf((*file), ",");
    putTreeFileHelper( file, bnode->rightPtr, bnode->rightDis );
    fprintf((*file), ")");
  }
  fprintf((*file),":%f", dis);
}
void bTreeNode::inOrderHelper( bTreeNode *bnode ) 
{
  if ( bnode != 0 ) {
    inOrderHelper(bnode->leftPtr);
    printf("| %.8f %4ld %4ld %.8f |\n", bnode->leftDis, bnode->id, bnode->numLeaves, bnode->rightDis);
    inOrderHelper(bnode->rightPtr);
  }
}
void bTreeNode::NilbTreeHelper( bTreeNode **bnode) 
{
  if ( *bnode != 0 ) {
    if ( (*bnode)->leftPtr != 0 ) {
      NilbTreeHelper( &((*bnode)->leftPtr) );
    }
    if ( (*bnode)->rightPtr != 0 ) {
      NilbTreeHelper( &((*bnode)->rightPtr) );
    }
    delete (*bnode);
    (*bnode) = 0;
  }
}
bTreeNode **bTreeNode::getLeafArray()
{
  bTreeNode **bnode, **bnodetemp;;
  bnode = new bTreeNode* [numLeaves+1];
  bnodetemp = bnode;
  bnodetemp++;
  getLeafArrayHelper(&(bnodetemp), this);
  return (bnode);
}
void bTreeNode::getLeafArrayHelper( bTreeNode*** barray, bTreeNode* bnode )
{
  if ( bnode->leftPtr != 0 && bnode->rightPtr != 0 ) {
    getLeafArrayHelper( barray, bnode->leftPtr );
    getLeafArrayHelper( barray, bnode->rightPtr );
  } else {
    (**barray) = bnode;
    ++(*barray);
  }
}
Int4 bTreeNode::getNumLeaves() 
{
  return ( numLeaves );
}
double bTreeNode::getDistance( bTreeNode *node2 ) 
{
  printf("WARNING: This function is not working correctly, and should not be used!\n");
  assert(0);

  Int4 i=0, j=0, totalLeaves=0;
  double node1Dis=0.0, node2Dis=0.0;
  bTreeNode **node1Array, **node2Array, *nodetemp;

  assert( parentPtr != 0 && node2->parentPtr != 0 );
  if ( this == node2 ) {
    return (0.0);
  }
  for ( nodetemp = this ; nodetemp->parentPtr != 0 ; nodetemp=nodetemp->parentPtr ) {}
  totalLeaves = nodetemp->numLeaves;

  node1Array = new bTreeNode*[totalLeaves];
  node2Array = new bTreeNode*[totalLeaves];
  nodetemp = this;
  do {
    ++i;
    if ( nodetemp->parentPtr->leftPtr == nodetemp ) {
//      printf("node1Dis before: %lf left\n",node1Dis);
      node1Dis += nodetemp->parentPtr->leftDis;
//      printf("node1Dis after:  %lf left\n",node1Dis);
    } else {
//      printf("node1Dis before: %lf right\n",node1Dis);
      node1Dis += nodetemp->parentPtr->rightDis;
//      printf("node1Dis after: %lf right\n",node1Dis);
    }
    nodetemp = nodetemp->parentPtr;
    node1Array[i] = nodetemp;
  } while ( nodetemp->parentPtr != 0 );
  nodetemp = node2;
  do {
    ++j;
    if ( j >= totalLeaves ) return(-2); //Added in case a bTreeNode ptr was passed that was in a larger, and different tree.
    if ( nodetemp->parentPtr->leftPtr == nodetemp ) {    	  
//      printf("node2Dis before: %lf left\n", node2Dis);
      node2Dis += nodetemp->parentPtr->leftDis;
//      printf("node2Dis after:  %lf left\n", node2Dis);
    } else {
//      printf("node2Dis before: %lf right\n", node2Dis);
      node2Dis += nodetemp->parentPtr->rightDis;
//      printf("node2Dis after:  %lf right\n", node2Dis);
    }
    nodetemp = nodetemp->parentPtr;
    node2Array[j] = nodetemp;
  } while ( nodetemp->parentPtr != 0 );  
  if ( node1Array[i] != node2Array[j] ) return(-3); // Check to make sure the root node is the same
  for(;;) {
    --i;
    --j;
    if ( node1Array[i] != node2Array[j] || node1Array[i] == 0 ) {
      return ( node1Dis + node2Dis );
    } else {
      if ( node1Array[i]->parentPtr->leftPtr == node1Array[i] ) {
//	printf("node1Dis before: %lf, node2Dis before: %lf left\n", node1Dis, node2Dis);
	node1Dis -= node1Array[i]->parentPtr->leftDis;
	node2Dis -= node2Array[j]->parentPtr->leftDis;
//	printf("node1Dis after: %lf, node2Dis after: %lf left\n", node1Dis, node2Dis);
      } else {
//	printf("node1Dis before: %lf, node2Dis before: %lf right\n", node1Dis, node2Dis);
	node1Dis -= node1Array[i]->parentPtr->rightDis;
	node2Dis -= node2Array[j]->parentPtr->rightDis;
//	printf("node1Dis after: %lf, node2Dis after: %lf right\n", node1Dis, node2Dis);
      }
    }
  }
}

