/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/**************
 bTreeNode.[ch]: Developed by Trevor Carlson <tcarlson@andrew.cmu.edu> and
                   Dr. Andrew Neuwald <neuwald@cshl.org>

     This class is a basic Binary Tree implementation used with the phylip_class package.
This package uses the (id:distance,id:distance); method to store binary tree's in file form.

  inOrderTraversal()     - A debugging function, prints out the tree "in-order".

  getTreeFile(FILE *)    - input/output of tree files.
  putTreeFile(FILE *)

  getNumLeaves()         - Returns the number of leaves below this node.
  getLeafArray()         - Returns an array of leaf nodes below this node. ( [1] through [numLeaves] )

  getDistance(bTreeNode*)- **NOT WORKING**
                         - This function will return the distance from one leaf to another.

Current Limitations:

     *****NOTE*****
     getDistance(bTreeNode *) is not functioning properly, and will produce incorrect
 results if used!

     Although this tree class can read and write files that are similar to those used by
 the phylip package, it cannot use them directly.  Currently, whilespace can ONLY be before
 the actual tree data, and not inside it.  

     I am using a special workaround to handle unrooted trees in this Binary Tree class.
 If rightDis == 0 of the "root" node, the tree will be considered unrooted and stored that 
 way. In a tree file, unrooted trees are stored as: 
 (id:distance, id:distance, id:distance ); 

     This tree class is not intended to be used for creating nodes with one leafPtr == 0.

 **************/
#ifndef __bTreeNode_h__
#define __bTreeNode_h__

#include <stdio.h>
#include <assert.h>
#include "phylip.h"

class bTreeNode {

 public:

  bTreeNode();
  bTreeNode(Int4);
  bTreeNode(bTreeNode*, bTreeNode*, double, double, Int4);
 ~bTreeNode();

void inOrderTraversal();
void getTreeFile(FILE *);
void putTreeFile(FILE *);
bTreeNode **getLeafArray();
Int4 getNumLeaves();
double getDistance( bTreeNode* );

 private:

void inOrderHelper( bTreeNode* );
double getTreeFileHelper( FILE **, bTreeNode** );
void putTreeFileHelper( FILE **, bTreeNode*, double );
void NilbTreeHelper( bTreeNode**); 
void getLeafArrayHelper( bTreeNode***, bTreeNode*);

  bTreeNode *leftPtr, *rightPtr, *parentPtr;
  double leftDis, rightDis;
  Int4 id, numLeaves;

};

#endif
