/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/* nbr_class.c 08/17/99 
   Modified by Trevor Carlson (tcarlson@andrew.cmu.edu)
   The Original program can be found at:
     http://evolution.genetics.washington.edu/phylip.html
*/

#include "nbr_class.h"

nbr_class::nbr_class() 
{
  compl_tree = false;
}
nbr_class::~nbr_class() {}
nbr_class::nbr_class( double **d, Int4 count) 
{   
  x = d;
  numsp = count;
  compl_tree = false;
}
tree_class &nbr_class::getTree() {

  if ( !compl_tree ) {
    startup();
    compl_tree = true;
  }
  return ( curtree );
}
void nbr_class::openfile(FILE **fp, char *filename, const char *mode ) {
  FILE *of;
  char file[100];
  strcpy(file,filename);
  while (1){
    of = fopen(file,mode);
    if (of)
      break;
    else {
      switch (*mode){
      case 'r':   

        printf("can't read %s\n",file);
	file[0] = '\0';
        while (file[0] =='\0'){
          printf("Please enter a new filename>");
          printf("this routine commented out.");
exit(1); // gets(file);
	} break;
      case 'w': case 'a':

        printf("can't write %s\n",file);

	file[0] = '\0';
        while (file[0] =='\0'){
          printf("Please enter a new filename>");
          printf("this routine commented out.");
exit(1); // gets(file);
	} break;
      }
    }
  }
  *fp=of;
}

double nbr_class::randum (void)

{ 
  Int4 i, j, k, sum;
  longer mult, newseed, seed;
  double v;

  for (i = 0; i <= 5; i++) seed[i] = 0;
  i = 0;
  do {
    seed[i] = inseed & 63; inseed /= 64; i++;
  } while (inseed != 0);

  mult[0] = 13;
  mult[1] = 24;
  mult[2] = 22;
  mult[3] = 6;
  for (i = 0; i <= 5; i++)
    newseed[i] = 0;
  for (i = 0; i <= 5; i++) {
    sum = newseed[i];
    k = i;
    if (i > 3)
      k = 3;
    for (j = 0; j <= k; j++)
      sum += mult[j] * seed[i - j];
    newseed[i] = sum;
    for (j = i; j <= 4; j++) {
      newseed[j + 1] += newseed[j] / 64;
      newseed[j] &= 63;
    }
  }
  memcpy(seed, newseed, sizeof(longer));
  seed[5] &= 3;
  v = 0.0;
  for (i = 0; i <= 5; i++)
    v = v / 64.0 + seed[i];
  v /= 4.0;
  return v;
}

void nbr_class::hookup(node_class *p, node_class *q){

 p->back = q; q->back = p;

}

void nbr_class::nodelabel(boolean isnode) {

if (isnode) printf("NODE"); else printf("OTU "); 

}
void nbr_class::jointree(void){
  /* calculate the tree */
  Int4 nc, nextnode, mini, minj, i, j, ia, ja, ii, jj, nude, iter;
  double diq, djq, dij, fotu2, total, tmin, dio, djo, bi, bj, bk, dmin, da;
  Int4 el[3];
  Vector av;
  intvector oc;

  for (i = 0; i <= numsp - 2; i++) {
    for (j = i + 1; j < numsp; j++) {
      da = (x[i][j] + x[j][i]) / 2.0;
      x[i][j] = da; x[j][i] = da;
    }
  }
  /* First initialization */
  fotu2 = numsp - 2.0;
  nextnode = numsp + 1;
  av =  new double[numsp];
  oc = new Int4[numsp];
  for (i = 0; i < numsp; i++) { av[i] = 0.0; oc[i] = 1; }
  /* Enter the main cycle */
  if (njoin) iter = numsp - 3;
  else iter = numsp - 1;
  for (nc = 1; nc <= iter; nc++) {
    for (j = 2; j <= numsp; j++) {
      for (i = 0; i <= j - 2; i++) x[j - 1][i] = x[i][j - 1];
    }
    tmin = 99999.0;
    /* Compute sij and minimize */
    for (ja = 2; ja <= numsp; ja++) {
      jj = enterorder[ja - 1];
      if (cluster[jj - 1] != NULL) {
        for (ia = 0; ia <= ja - 2; ia++) {
          ii = enterorder[ia];
          if (cluster[ii - 1] != NULL) {
            if (njoin) {
              diq = 0.0; djq = 0.0; dij = x[ii - 1][jj - 1];
              for (i = 0; i < numsp; i++) {
                diq += x[i][ii - 1]; djq += x[i][jj - 1];
              }
              total = fotu2 * dij - diq - djq;
            } else total = x[ii - 1][jj - 1];
            if (total < tmin) { tmin = total; mini = ii; minj = jj; }
          }
        }
      }
    }
    /* compute lengths and print */
    if (njoin) {
      dio = 0.0; djo = 0.0;
      for (i = 0; i < numsp; i++) {
        dio += x[i][mini - 1]; djo += x[i][minj - 1];
      }
      dmin = x[mini - 1][minj - 1];
      dio = (dio - dmin) / fotu2; djo = (djo - dmin) / fotu2;
      bi = (dmin + dio - djo) * 0.5; bj = dmin - bi;
      bi -= av[mini - 1]; bj -= av[minj - 1];
    } else {
      bi = x[mini - 1][minj - 1] / 2.0 - av[mini - 1];
      bj = x[mini - 1][minj - 1] / 2.0 - av[minj - 1];
      av[mini - 1] += bi;
    }
    
    if (progress) {
      printf("CYCLE %3ld: ", iter - nc + 1);
      if (njoin) nodelabel(av[mini - 1] > 0.0);
      else nodelabel(oc[mini - 1] > 1.0);
      printf("%3ld (%10.5f) JOINS ", mini, bi);
      if (njoin) nodelabel(av[minj - 1] > 0.0);
      else nodelabel(oc[minj - 1] > 1.0);
      printf("%3ld (%10.5f)\n", minj, bj);
    }
    
    hookup(curtree.nodep[nextnode - 1]->next, cluster[mini - 1]);
    hookup(curtree.nodep[nextnode - 1]->next->next, cluster[minj - 1]);
    cluster[mini - 1]->v = bi; cluster[minj - 1]->v = bj;
    cluster[mini - 1]->back->v = bi; cluster[minj - 1]->back->v = bj;
    cluster[mini - 1] = curtree.nodep[nextnode - 1];
    cluster[minj - 1] = NULL; nextnode++;
    if (njoin) av[mini - 1] = dmin * 0.5;
    /* re-initialization */
    fotu2 -= 1.0;
    for (j = 0; j < numsp; j++) {
      if (cluster[j] != NULL) {
        if (njoin) {
          da = (x[mini - 1][j] + x[minj - 1][j]) * 0.5;
          if (mini - j - 1 < 0) x[mini - 1][j] = da;
          if (mini - j - 1 > 0) x[j][mini - 1] = da;
        } else {
          da = x[mini - 1][j] * oc[mini - 1] + x[minj - 1][j] * oc[minj - 1];
          da /= oc[mini - 1] + oc[minj - 1];
          x[mini - 1][j] = da; x[j][mini - 1] = da;
        }
      }
    }
    for (j = 0; j < numsp; j++) { x[minj - 1][j] = 0.0; x[j][minj - 1] = 0.0; }
    oc[mini - 1] += oc[minj - 1];
  }
  /* the last cycle */
  nude = 1;
  for (i = 1; i <= numsp; i++) {
    if (cluster[i - 1] != NULL) { el[nude - 1] = i; nude++; }
  }
  if (!njoin) { curtree.start = cluster[el[0] - 1]; return; }
  bi = (x[el[0] - 1][el[1] - 1] + x[el[0] - 1][el[2] - 1] - x[el[1] - 1]
        [el[2] - 1]) * 0.5;
  bj = x[el[0] - 1][el[1] - 1] - bi;
  bk = x[el[0] - 1][el[2] - 1] - bi;
  bi -= av[el[0] - 1]; bj -= av[el[1] - 1]; bk -= av[el[2] - 1];
  
  if (progress) {
    printf("LAST CYCLE:\n"); putchar(' ');
    nodelabel(av[el[0] - 1] > 0.0);
    printf("%3ld  (%10.5f) JOINS ", el[0], bi);
    nodelabel(av[el[1] - 1] > 0.0);
    printf("%3ld  (%10.5f) JOINS ", el[1], bj);
    nodelabel(av[el[2] - 1] > 0.0);
    printf("%3ld  (%10.5f)\n", el[2], bk);
  }
  
  hookup(curtree.nodep[nextnode - 1], cluster[el[0] - 1]);
  hookup(curtree.nodep[nextnode - 1]->next, cluster[el[1] - 1]);
  hookup(curtree.nodep[nextnode - 1]->next->next, cluster[el[2] - 1]);
  cluster[el[0] - 1]->v = bi;
  cluster[el[1] - 1]->v = bj;
  cluster[el[2] - 1]->v = bk;
  cluster[el[0] - 1]->back->v = bi;
  cluster[el[1] - 1]->back->v = bj;
  cluster[el[2] - 1]->back->v = bk;
  curtree.start = cluster[el[0] - 1]->back;
  delete [] av; delete [] oc;

}
void nbr_class::maketree(void){
  /* construct the tree */
  Int4 i, j, k;

  /* Two data elements for maketree(), jointree() */
  enterorder = new Int4[numsp];
  cluster = new node_class * [numsp];

  if (progress) putchar('\n');

  curtree.setupTree( numsp, njoin);

  for (i = 1; i <= numsp; i++) enterorder[i - 1] = i;
  if (jumble) {
    for (i = 0; i < numsp; i++) {
      j = (Int4)(randum() * numsp) + 1;
      k = enterorder[j - 1];
      enterorder[j - 1] = enterorder[i];
      enterorder[i] = k;
    }
  }
  for (i = 0; i < numsp; i++) cluster[i] = curtree.nodep[i];
  jointree();
  if (outgropt) curtree.start = curtree.nodep[outgrno - 1]->back;

  /* Removed treeout. treeprint, now in another class,        */
  /* Calls are now made to tree_class.c                       */

  delete [] enterorder;
  delete [] cluster;
}

void nbr_class::startup(void) {

  /* Variable Setup */  

  jumble = false;
  inseed = 364379;

  lower = false;
  upper = false;

  outgrno = 1;
  outgropt = false;

  trout = false;
  printdata = false;
  progress = false;

  njoin = true;

  Int4 i;
  /* END variable setup */

  if ( printdata) {
    openfile(&outfile,outfilename,"w");

    putc('\n', outfile);
    if (njoin) fprintf(outfile," Neighbor-joining method\n");
    else fprintf(outfile," UPGMA method\n");
    fprintf(outfile,"\n Negative branch lengths allowed\n\n");

    Int4 i, j, k, columns;
    columns = 6;

    if (printdata) {

      fprintf(outfile, "\nName                       Distances");
      fprintf(outfile, "\n----                       ---------");
      fprintf(outfile, "\n\n");

      for (i = 0; i < numsp; i++) {
	fprintf( outfile, "%10ld", i+1);
	/*
	for (j = 0; j < namelength; j++) putc(names[i][j], outfile);
	*/
	putc(' ', outfile);
	for (j = 1; j <= numsp; j++) {
	  fprintf(outfile, "%10.5f", x[i][j - 1]);
	  if (j % columns == 0 && j < numsp) {
	    putc('\n', outfile);
	    for (k = 1; k <= 11; k++) putc(' ', outfile);
	    //for (k = 1; k <= namelength + 1; k++) putc(' ', outfile);
	  }
	}
	putc('\n', outfile);
      }
      putc('\n', outfile); 
    }
    FClose(outfile); 
  }

  maketree();


}


