/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/* nbr_class.h 08/17/99 
   Modified by Trevor Carlson (tcarlson@andrew.cmu.edu)
   The Original program can be found at:
     http://evolution.genetics.washington.edu/phylip.html

   nbr_class: This class takes a distance matrix < double** > and creates a tree from it.
  Currently, the tree that is passed out IS NOT duplicated, and is the one that was used 
  internally.  Also, it will not be destroyed.

  << See the tree_class.h file for more warnings about the tree_class >>

   nbr_class( double **, Int4 ) - pass in the distance matrix, and the number of proteins here.
   getTree()                    - retrieves the tree_class object that is used internally.


        TODO: it might be more useful to pass in the values with the accessor instead of
   at the constructor.

        BUGS: only the tree_class problem is currently known.
*/
/* version 3.52c. (c) Copyright 1993 by Joseph Felsenstein and Mary Kuhner.
   Written by Mary Kuhner, Jon Yamato, Joseph Felsenstein, Akiko Fuseki,
   Sean Lamont, and Andrew Keeffe.
   Permission is granted to copy and use this program provided no fee is
   charged for it and provided that this copyright notice is not removed. */
#ifndef __nbr_class_h__
#define __nbr_class_h__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "tree_class.h"

typedef unsigned char boolean;
typedef double *Vector;
typedef Int4 *intvector;
typedef Int4 longer[6];

class nbr_class {
 public:

  nbr_class();
 ~nbr_class();
  nbr_class( double **, Int4 );

  tree_class &getTree();

 private:

  FILE *outfile;
  char *outfilename;
  char *nbroutfilename;

  Int4 numsp, inseed, outgrno;
  Vector *x, *y;
  boolean jumble, lower, upper, outgropt, trout,
               printdata, progress, treeprint, njoin;
  boolean compl_tree; /* Added to check if tree has been completed */
  tree_class curtree;
  Int4 *enterorder;
  node_class **cluster;

  void openfile(FILE **fp, char *filename, const char *mode );
  double randum (void);
  void hookup(node_class *p, node_class *q);
  void nodelabel(boolean isnode);
  void jointree(void);
  void maketree(void);
  void startup(void);

 };

#endif

