/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "pdis_class.h"

pdis_class::pdis_class() {}

pdis_class::pdis_class( unsigned char **seq, Int4   numsp, Int4  length ) 
{
  assert ( seq != NULL || numsp < 2 || length < 1 );

  Int4 i, j;

  spp = numsp;
  chars = length;

  compdist = false;

  setupGlobals();
  
  for ( i = 0 ; i < spp ; i++ ) {
    for ( j = 0 ; j < chars ; j++ ) {
      node[i][j] = convdata ( seq[i][j] ) ;
    }
  }
}

pdis_class::~pdis_class() 
{
  distroyVars();
}
double** pdis_class::getDistance ()
{
  int i, j;
  
  startup();
  
  double **distance = new double*[spp+1];
  for ( i = 1 ; i <= spp ; i++ ) {
    distance[i] = new double[spp];
  }

  for ( i = 1 ; i <=spp ; i++ ) {
    for ( j = 0 ; j < spp ; j++ ) {
      distance[i][j] = d[i-1][j];
    }
  }
  
  //Below is for outputting the lower half
  //Remember to allocate less memory in the above statment!
  /*for ( i = 2 ; i <= spp ; i++ ) {
    for ( j = 0 ; i > j + 1 ; j++ ) {
      distance[i][j] = d[i-1][j];
    }
  }*/
  return distance; 
}
void pdis_class::distroyVars()
{
  Int4 i;
  for (i = 0; i < spp; i++) {
    delete [] node[i];
    delete [] d[i];
  }  
  delete [] node;
  delete [] d;
}
void pdis_class::openfile(FILE **fp, char *filename, const char *mode, char *perm )
{
  FILE *of;
  char file[100];
  strcpy(file,filename);
  while (1){
    of = fopen(file,mode);
    if (of)
      break;
    else {
      switch (*mode){
      case 'r':
        fprintf(stderr,"Can't read %s\n",file);
	
        file[0]='\0';
        while (file[0] =='\0'){
          fprintf(stderr,"Please enter a new filename>");
exit(1); // gets(file);
	} break;
      case 'w':
        fprintf(stderr,"Can't write %s\n",file);
        file[0] = '\0';
        while (file[0] =='\0'){
          fprintf(stderr,"Please enter a new filename>");
exit(1); // gets(file);
	} break;
      }
    }
  }
  *fp=of;
  if (perm != NULL)
    strcpy(perm,file);
}
void pdis_class::printHeadings() {

  Int4 i, j, k, l;
  char charstate;

//  if (printdata)
    fprintf(outfile, "%2ld species, %3ld  sites\n\n", spp, chars);

//  if (printdata)
    fprintf(outfile, "\nProtein distance algorithm, version 3.5\n\n");
  //putchar('\n');

  /* if (progress)
    putchar('\n');
   */

  
  j = nmlngth + (chars + (chars - 1) / 10) / 2 - 5;
  if (j < nmlngth - 1)
    j = nmlngth - 1;
  if (j > 37)
    j = 37;
    
//  if (printdata) {
    fprintf(outfile, "Name");
    for (i = 1; i <= j; i++)
      putc(' ', outfile);
    fprintf(outfile, "Sequences\n");
    fprintf(outfile, "----");
    for (i = 1; i <= j; i++)
      putc(' ', outfile);
    fprintf(outfile, "---------\n\n");
//  }

//  if ( printdata) {
    for (i = 1; i <= ((chars - 1) / 60 + 1); i++) {
      for (j = 1; j <= spp; j++) {
	fprintf(outfile, "%10d", j); /* Inserted instead of names */
	fprintf(outfile, "   ");
	l = i * 60;
	if (l > chars)
	  l = chars;
	for (k = (i - 1) * 60 + 1; k <= l; k++) {
	  if (j > 1 && node[j - 1][k - 1] == node[0][k - 1])
	    charstate = '.';
	  else {
	    switch (node[j - 1][k - 1]) {

	    case ala: charstate = 'A'; break;
	    case asx: charstate = 'B'; break;
	    case cys: charstate = 'C'; break;
	    case asp: charstate = 'D'; break;
	    case glu: charstate = 'E'; break;
	    case phe: charstate = 'F'; break;
	    case gly: charstate = 'G'; break;
	    case his: charstate = 'H'; break;
	    case ileu: charstate = 'I'; break;
	    case lys: charstate = 'K'; break;
	    case leu: charstate = 'L'; break;
	    case met: charstate = 'M'; break;
	    case asn: charstate = 'N'; break;
	    case pro: charstate = 'P'; break;
	    case gln: charstate = 'Q'; break;
	    case arg: charstate = 'R'; break;
	    case ser: charstate = 'S'; break;
	    case thr: charstate = 'T'; break;
	    case val: charstate = 'V'; break;
	    case trp: charstate = 'W'; break;
	    case tyr: charstate = 'Y'; break;
	    case glx: charstate = 'Z'; break;
	    case del: charstate = '-'; break;
	    case stop: charstate = '*'; break;
	    case unk: charstate = 'X'; break;
	    case quest: charstate = '?'; break;
	    }
	  }
	  putc(charstate, outfile);
	  if (k % 10 == 0 && k % 60 != 0)
	    putc(' ', outfile);
	}
	putc('\n', outfile);
      }
      putc('\n', outfile);
    }
    putc('\n', outfile);
//  }
  //if (printdata)
    putc('\n', outfile);
}
void pdis_class::setupGlobals() {

  Int4 i;

  /* Setup Global Default variables */
  mulsets = false;
  datasets = 1;

  printdis = false;
  printdata = false;
  progress = true;
  interleaved = true;
  ttratio = 2.0;
  whichcode = universal;
  whichcat = george;
  basesequal = true;
  freqa = 0.25;
  freqc = 0.25;
  freqg = 0.25;
  freqt = 0.25;
  usepam = true;
  kimura = false;
  ease = 0.457;

  node = new aas*[spp];
  for (i = 0; i < spp; i++)
    node[i] = new aas[chars];

  d      = new double *[spp];

  for (i=0;i<spp;++i){
    d[i] = new double[spp];
  }

  /* calculations related to transition-transversion ratio */
  double aa, bb, freqr, freqy, freqgr, freqty;

  freqr = freqa + freqg;
  freqy = freqc + freqt;
  freqgr = freqg / freqr;
  freqty = freqt / freqy;
  aa = ttratio * freqr * freqy - freqa * freqg - freqc * freqt;
  bb = freqa * freqgr + freqc * freqty;
  xi = aa / (aa + bb);
  xv = 1.0 - xi;
  if (xi <= 0.0 && xi >= -epsilon)
    xi = 0.0;
  if (xi < 0.0){
    fprintf(stderr,"THIS TRANSITION-TRANSVERSION RATIO IS IMPOSSIBLE WITH");
    fprintf(stderr," THESE BASE FREQUENCIES\n");
    exit(-1);
  }
  /* Setup concerning  */
  if (!kimura)
    code();
  if (!(usepam || kimura)) {
    cats();
    maketrans();
    qreigen(prob, 20L);
  } else {
    if (kimura)
      fracchange = 1.0;
    else
      pameigen();
  }
}


aas  pdis_class::convdata( char charstate )
{
  aas aa;   /* temporary amino acid for input */

  /* Added to convert data to internally used data.. */
          switch (charstate) {

          case 'A':
            aa = ala;
            break;

          case 'B':
            aa = asx;
            break;

          case 'C':
            aa = cys;
            break;

          case 'D':
            aa = asp;
            break;

          case 'E':
            aa = glu;
            break;

          case 'F':
            aa = phe;
            break;

          case 'G':
            aa = gly;
            break;

          case 'H':
            aa = his;
            break;

          case 'I':
            aa = ileu;
            break;

          case 'K':
            aa = lys;
            break;

          case 'L':
            aa = leu;
            break;

          case 'M':
            aa = met;
            break;

          case 'N':
            aa = asn;
            break;

          case 'P':
            aa = pro;
            break;

          case 'Q':
            aa = gln;
            break;
	    
          case 'R':
            aa = arg;
            break;

          case 'S':
            aa = ser;
            break;

          case 'T':
            aa = thr;
            break;

          case 'V':
            aa = val;
            break;

          case 'W':
            aa = trp;
            break;

          case 'X':
            aa = unk;
            break;

          case 'Y':
            aa = tyr;
            break;

          case 'Z':
            aa = glx;
            break;

          case '*':
            aa = stop;
            break;

          case '?':
            aa = quest;
            break;

          case '-':
            aa = del;
            break;

	  default:
	    fprintf(stderr,"\nInvalid character in char array, assuming unknown character.");
	    aa = unk;
	    break;
          }
          return ( aa );
}
void pdis_class::code()
{
  /* make up table of the code 1 = u, 2 = c, 3 = a, 4 = g */
  Int4 n;
  aas b;

  trans[0][0][0] = phe;
  trans[0][0][1] = phe;
  trans[0][0][2] = leu;
  trans[0][0][3] = leu;
  trans[0][1][0] = ser;
  trans[0][1][1] = ser;
  trans[0][1][2] = ser;
  trans[0][1][3] = ser;
  trans[0][2][0] = tyr;
  trans[0][2][1] = tyr;
  trans[0][2][2] = stop;
  trans[0][2][3] = stop;
  trans[0][3][0] = cys;
  trans[0][3][1] = cys;
  trans[0][3][2] = stop;
  trans[0][3][3] = trp;
  trans[1][0][0] = leu;
  trans[1][0][1] = leu;
  trans[1][0][2] = leu;
  trans[1][0][3] = leu;
  trans[1][1][0] = pro;
  trans[1][1][1] = pro;
  trans[1][1][2] = pro;
  trans[1][1][3] = pro;
  trans[1][2][0] = his;
  trans[1][2][1] = his;
  trans[1][2][2] = gln;
  trans[1][2][3] = gln;
  trans[1][3][0] = arg;
  trans[1][3][1] = arg;
  trans[1][3][2] = arg;
  trans[1][3][3] = arg;
  trans[2][0][0] = ileu;
  trans[2][0][1] = ileu;
  trans[2][0][2] = ileu;
  trans[2][0][3] = met;
  trans[2][1][0] = thr;
  trans[2][1][1] = thr;
  trans[2][1][2] = thr;
  trans[2][1][3] = thr;
  trans[2][2][0] = asn;
  trans[2][2][1] = asn;
  trans[2][2][2] = lys;
  trans[2][2][3] = lys;
  trans[2][3][0] = ser;
  trans[2][3][1] = ser;
  trans[2][3][2] = arg;
  trans[2][3][3] = arg;
  trans[3][0][0] = val;
  trans[3][0][1] = val;
  trans[3][0][2] = val;
  trans[3][0][3] = val;
  trans[3][1][0] = ala;
  trans[3][1][1] = ala;
  trans[3][1][2] = ala;
  trans[3][1][3] = ala;
  trans[3][2][0] = asp;
  trans[3][2][1] = asp;
  trans[3][2][2] = glu;
  trans[3][2][3] = glu;
  trans[3][3][0] = gly;
  trans[3][3][1] = gly;
  trans[3][3][2] = gly;
  trans[3][3][3] = gly;
  if (whichcode == mito)
    trans[0][3][2] = trp;
  if (whichcode == vertmito) {
    trans[0][3][2] = trp;
    trans[2][3][2] = stop;
    trans[2][3][3] = stop;
    trans[2][0][2] = met;
  }
  if (whichcode == flymito) {
    trans[0][3][2] = trp;
    trans[2][0][2] = met;
    trans[2][3][2] = ser;
  }
  if (whichcode == yeastmito) {
    trans[0][3][2] = trp;
    trans[1][0][2] = thr;
    trans[2][0][2] = met;
  }
  n = 0;
  for (b = ala; (Int4)b <= (Int4)val; b = (aas)((Int4)b + 1)) {
    n++;
    numaa[(Int4)b - (Int4)ala] = n;
  }
}
void pdis_class::cats()
{
  /* define categories of amino acids */
  aas b;
  /* fundamental subgroups */
  cat[(Int4)cys - (Int4)ala] = 1;
  cat[(Int4)met - (Int4)ala] = 2;
  cat[(Int4)val - (Int4)ala] = 3;
  cat[(Int4)leu - (Int4)ala] = 3;
  cat[(Int4)ileu - (Int4)ala] = 3;
  cat[(Int4)gly - (Int4)ala] = 4;
  cat[0] = 4;
  cat[(Int4)ser - (Int4)ala] = 4;
  cat[(Int4)thr - (Int4)ala] = 4;
  cat[(Int4)pro - (Int4)ala] = 5;
  cat[(Int4)phe - (Int4)ala] = 6;
  cat[(Int4)tyr - (Int4)ala] = 6;
  cat[(Int4)trp - (Int4)ala] = 6;
  cat[(Int4)glu - (Int4)ala] = 7;
  cat[(Int4)gln - (Int4)ala] = 7;
  cat[(Int4)asp - (Int4)ala] = 7;
  cat[(Int4)asn - (Int4)ala] = 7;
  cat[(Int4)lys - (Int4)ala] = 8;
  cat[(Int4)arg - (Int4)ala] = 8;
  cat[(Int4)his - (Int4)ala] = 8;
  if (whichcat == george) {
    /* George, Hunt and Barker: sulfhydryl, small hydrophobic, small hydrophilic,
                              aromatic, acid/acid-amide/hydrophilic, basic */
    for (b = ala; (Int4)b <= (Int4)val; b = (aas)((Int4)b + 1)) {
      if (cat[(Int4)b - (Int4)ala] == 3)
	cat[(Int4)b - (Int4)ala] = 2;
      if (cat[(Int4)b - (Int4)ala] == 5)
	cat[(Int4)b - (Int4)ala] = 4;
    }
  }
  if (whichcat == chemical) {
    /* Conn and Stumpf:  monoamino, aliphatic, heterocyclic, aromatic, dicarboxylic,
                       basic */
    for (b = ala; (Int4)b <= (Int4)val; b = (aas)((Int4)b + 1)) {
      if (cat[(Int4)b - (Int4)ala] == 2)
	cat[(Int4)b - (Int4)ala] = 1;
      if (cat[(Int4)b - (Int4)ala] == 4)
	cat[(Int4)b - (Int4)ala] = 3;
    }
  }
  /* Ben Hall's personal opinion */
  if (whichcat != hall)
    return;
  for (b = ala; (Int4)b <= (Int4)val; b = (aas)((Int4)b + 1)) {
    if (cat[(Int4)b - (Int4)ala] == 3)
      cat[(Int4)b - (Int4)ala] = 2;
  }
}
void pdis_class::maketrans()
{
  /* Make up transition probability matrix from code and category tables */
  Int4 i, j, k, m, n, s, nb1, nb2;
  double x, sum;
  Int4 sub[3], newsub[3];
  double f[4], g[4];
  aas b1, b2;
  double TEMP, TEMP1, TEMP2, TEMP3;

  for (i = 0; i <= 19; i++) {
    pie[i] = 0.0;
    for (j = 0; j <= 19; j++)
      prob[i][j] = 0.0;
  }
  f[0] = freqt;
  f[1] = freqc;
  f[2] = freqa;
  f[3] = freqg;
  g[0] = freqc + freqt;
  g[1] = freqc + freqt;
  g[2] = freqa + freqg;
  g[3] = freqa + freqg;
  TEMP = f[0];
  TEMP1 = f[1];
  TEMP2 = f[2];
  TEMP3 = f[3];
  fracchange = xi * (2 * f[0] * f[1] / g[0] + 2 * f[2] * f[3] / g[2]) +
      xv * (1 - TEMP * TEMP - TEMP1 * TEMP1 - TEMP2 * TEMP2 - TEMP3 * TEMP3);
  sum = 0.0;
  for (i = 0; i <= 3; i++) {
    for (j = 0; j <= 3; j++) {
      for (k = 0; k <= 3; k++) {
	if (trans[i][j][k] != stop)
	  sum += f[i] * f[j] * f[k];
      }
    }
  }
  for (i = 0; i <= 3; i++) {
    sub[0] = i + 1;
    for (j = 0; j <= 3; j++) {
      sub[1] = j + 1;
      for (k = 0; k <= 3; k++) {
	sub[2] = k + 1;
	b1 = trans[i][j][k];
	for (m = 0; m <= 2; m++) {
	  s = sub[m];
	  for (n = 1; n <= 4; n++) {
	    memcpy(newsub, sub, sizeof(Int4) * 3L);
	    newsub[m] = n;
	    x = f[i] * f[j] * f[k] / (3.0 * sum);
	    if (((s == 1 || s == 2) && (n == 3 || n == 4)) ||
		((n == 1 || n == 2) && (s == 3 || s == 4)))
	      x *= xv * f[n - 1];
	    else
	      x *= xi * f[n - 1] / g[n - 1] + xv * f[n - 1];
	    b2 = trans[newsub[0] - 1][newsub[1] - 1][newsub[2] - 1];
	    if (b1 != stop) {
	      nb1 = numaa[(Int4)b1 - (Int4)ala];
	      pie[nb1 - 1] += x;
	      if (b2 != stop) {
		nb2 = numaa[(Int4)b2 - (Int4)ala];
		if (cat[(Int4)b1 - (Int4)ala] != cat[(Int4)b2 - (Int4)ala]) {
		  prob[nb1 - 1][nb2 - 1] += x * ease;
		  prob[nb1 - 1][nb1 - 1] += x * (1.0 - ease);
		} else
		  prob[nb1 - 1][nb2 - 1] += x;
	      } else
		prob[nb1 - 1][nb1 - 1] += x;
	    }
	  }
	}
      }
    }
  }
  for (i = 0; i <= 19; i++)
    prob[i][i] -= pie[i];
  for (i = 0; i <= 19; i++) {
    for (j = 0; j <= 19; j++)
      prob[i][j] /= sqrt(pie[i] * pie[j]);
  }
  /* computes pi^(1/2)*B*pi^(-1/2)  */
}
void pdis_class::givens(double (*a)[20],
	    Int4 i, 
	    Int4 j, 
	    Int4 n,
	    double ctheta, 
	    double stheta,
	    boolean left )
{
  /* Givens transform at i,j for 1..n with angle theta */
  Int4 k;
  double v;

  for (k = 0; k < n; k++) {
    if (left) {
      v = ctheta * a[i - 1][k] + stheta * a[j - 1][k];
      a[j - 1][k] = ctheta * a[j - 1][k] - stheta * a[i - 1][k];
      a[i - 1][k] = v;
    } else {
      v = ctheta * a[k][i - 1] + stheta * a[k][j - 1];
      a[k][j - 1] = ctheta * a[k][j - 1] - stheta * a[k][i - 1];
      a[k][i - 1] = v;
    }
  }
}
void pdis_class::coeffs(double x, 
	    double y, 
	    double *c, 
	    double *s,
	    double accuracy)
{
  /* compute cosine and sine of theta */
  double root;

  root = sqrt(x * x + y * y);
  if (root < accuracy) {
    *c = 1.0;
    *s = 0.0;
  } else {
    *c = x / root;
    *s = y / root;
  }
}
void pdis_class::tridiag(double (*a)[20],
	     Int4 n,
	     double accuracy)
{
  /* Givens tridiagonalization */
  Int4 i, j;
  double s, c;

  for (i = 2; i < n; i++) {
    for (j = i + 1; j <= n; j++) {
      coeffs(a[i - 2][i - 1], a[i - 2][j - 1], &c, &s,accuracy);
      givens(a, i, j, n, c, s, true);
      givens(a, i, j, n, c, s, false);
      givens(eigvecs, i, j, n, c, s, true);
    }
  }
}
void pdis_class::shiftqr(double (*a)[20],
	     Int4 n,
	     double accuracy)
{
  /* QR eigenvalue-finder */
  Int4 i, j;
  double approx, s, c, v, TEMP, TEMP1;

  for (i = n; i >= 2; i--) {
    do {
      TEMP = a[i - 2][i - 2] - a[i - 1][i - 1];
      TEMP1 = a[i - 1][i - 2];
      v = sqrt(TEMP * TEMP + TEMP1 * TEMP1);
      approx = a[i - 2][i - 2] + a[i - 1][i - 1];
      if (a[i - 1][i - 1] < a[i - 2][i - 2])
	approx = (approx - v) / 2.0;
      else
	approx = (approx + v) / 2.0;
      for (j = 0; j < i; j++)
	a[j][j] -= approx;
      for (j = 1; j < i; j++) {
	coeffs(a[j - 1][j - 1], a[j][j - 1], &c, &s, accuracy);
	givens(a, j, j + 1, i, c, s, true);
	givens(a, j, j + 1, i, c, s, false);
	givens(eigvecs, j, j + 1, n, c, s, true);
      }
      for (j = 0; j < i; j++)
	a[j][j] += approx;
    } while (fabs(a[i - 1][i - 2]) > accuracy);
  }
}
void pdis_class::qreigen(double (*prob_var)[20],
			 Int4 n)
{
  /* QR eigenvector/eigenvalue method for symmetric matrix */
  double accuracy;
  Int4 i, j;

  accuracy = 1.0e-6;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++)
      eigvecs[i][j] = 0.0;
    eigvecs[i][i] = 1.0;
  }
  tridiag(prob_var, n, accuracy);
  shiftqr(prob_var, n, accuracy);
  for (i = 0; i < n; i++)
    eig[i] = prob_var[i][i];
  for (i = 0; i <= 19; i++) {
    for (j = 0; j <= 19; j++)
      prob_var[i][j] = sqrt(pie[j]) * eigvecs[i][j];
  }
  /* prob[i][j] is the value of U' times pi^(1/2) */
}
void pdis_class::pameigen()
{
  /* eigenanalysis for PAM matrix, precomputed */
  memcpy(prob,pamprobs,sizeof(pamprobs));
  memcpy(eig,pameigs,sizeof(pameigs));
  fracchange = 0.01;
}
void pdis_class::predict(Int4 nb1,
			 Int4 nb2)
{
  /* make contribution to prediction of this aa pair */
  Int4 m;
  double TEMP;


  for (m = 0; m <= 19; m++) {
    elambdat = exp(tt * eig[m]);
    q = prob[m][nb1 - 1] * prob[m][nb2 - 1] * elambdat;
    p += q;
    dp += eig[m] * q;
    TEMP = eig[m];
    d2p += TEMP * TEMP * q;
  }

}
void pdis_class::makedists()
{
  /* compute the distances */
  Int4 i, j, k, m, n, iterations, nb1, nb2;
  double delta, lnlike, slope, curv;
  boolean neginfinity, inf;
  aas b1, b2;
  if (printdis) {
    fprintf(outfile, "%5ld\n", spp);
  }
  if (progress)
    fprintf(stderr,"Computing distances:\n");
  for (i = 1; i <= spp; i++) {
    if (progress)
      fprintf(stderr,"  ");
    if (progress) {
      fprintf(stderr, "%10d", i);
    }
    if (progress)
      fprintf(stderr,"   ");
    d[i - 1][i - 1] = 0.0;
    for (j = 0; j <= i - 2; j++) {
      if (!kimura) {
	if (usepam)
	  tt = 10.0;
	else
	  tt = 1.0;
	delta = tt / 2.0;
	iterations = 0;
        inf = false;
	do {
	  lnlike = 0.0;
	  slope = 0.0;
	  curv = 0.0;
	  neginfinity = false;
	  for (k = 0; k < chars; k++) {
	    b1 = node[i - 1][k];
	    b2 = node[j][k];
	    if (b1 != stop && b1 != del && b1 != quest && b1 != unk &&
		b2 != stop && b2 != del && b2 != quest && b2 != unk) {
	      p = 0.0;
	      dp = 0.0;
	      d2p = 0.0;
	      nb1 = numaa[(Int4)b1 - (Int4)ala];
	      nb2 = numaa[(Int4)b2 - (Int4)ala];
	      if (b1 != asx && b1 != glx && b2 != asx && b2 != glx)
		predict(nb1, nb2);
	      else {
		if (b1 == asx) {
		  if (b2 == asx) {
		    predict(3L, 3L);
		    predict(3L, 4L);
		    predict(4L, 3L);
		    predict(4L, 4L);
		  } else {
		    if (b2 == glx) {
		      predict(3L, 6L);
		      predict(3L, 7L);
		      predict(4L, 6L);
		      predict(4L, 7L);
		    } else {
		      predict(3L, nb2);
		      predict(4L, nb2);
		    }
		  }
		} else {
		  if (b1 == glx) {
		    if (b2 == asx) {
		      predict(6L, 3L);
		      predict(6L, 4L);
		      predict(7L, 3L);
		      predict(7L, 4L);
		    } else {
		      if (b2 == glx) {
			predict(6L, 6L);
			predict(6L, 7L);
			predict(7L, 6L);
			predict(7L, 7L);
		      } else {
			predict(6L, nb2);
			predict(7L, nb2);
		      }
		    }
		  } else {
		    if (b2 == asx) {
		      predict(nb1, 3L);
		      predict(nb1, 4L);
		      predict(nb1, 3L);
		      predict(nb1, 4L);
		    } else if (b2 == glx) {
		      predict(nb1, 6L);
		      predict(nb1, 7L);
		      predict(nb1, 6L);
		      predict(nb1, 7L);
		    }
		  }
		}
	      }
	      if (p <= 0.0)
		neginfinity = true;
	      else {
		lnlike += log(p);
		slope += dp / p;
		curv += d2p / p - dp * dp / (p * p);
	      }
	    }
	  }
	  iterations++;
	  if (!neginfinity) {
	    if (curv < 0.0) {
	      tt -= slope / curv;
              if (tt > 10000.0) {
                fprintf(stderr,"\nWARNING: INFINITE DISTANCE BETWEEN SPECIES %d AND %d; -1.0 WAS WRITTEN\n", i, j);
                tt = -1.0/fracchange;
                inf = true;
                iterations = 20;
              }
            }
	    else {
	      if ((slope > 0.0 && delta < 0.0) || (slope < 0.0 && delta > 0.0))
		delta /= -2;
	      tt += delta;
	    }
	  } else {
	    delta /= -2;
	    tt += delta;
	  }
	  if (tt < epsilon && !inf)
	    tt = epsilon;
	} while (iterations != 20);
      } else {
	m = 0;
	n = 0;
	for (k = 0; k < chars; k++) {
	  b1 = node[i - 1][k];
	  b2 = node[j][k];
	  if ((Int4)b1 <= (Int4)val && (Int4)b2 <= (Int4)val) {
	    if (b1 == b2)
	      m++;
	    n++;
	  }
	}
	p = 1 - (double)m / n;
	dp = 1.0 - p - 0.2 * p * p;
	if (dp < 0.0) {
	  fprintf(stderr,
	    "\nDISTANCE BETWEEN SEQUENCES %3ld AND %3ld IS TOO LARGE FOR KIMURA FORMULA\n",
	    i, j + 1);
	  tt = -1.0;
	} else
	  tt = -log(dp);
      }
      d[i - 1][j] = fracchange * tt;
      d[j][i - 1] = d[i - 1][j];
      if (progress)
	//putchar('.'); // going from buffered to unbuffered, is it significant?
        putc('.', stderr);
    }
    if (progress)
      //putchar('\n');
      putc('\n', stderr);
  }
  if (printdis) {
    for (i = 0; i < spp; i++) {
      fprintf( outfile, "%10d", i+1 );
      fprintf(outfile, "   ");
      for (j = 0; j < spp; j++)
	fprintf(outfile, "%9.5f", d[i][j]);
      putc('\n', outfile);
    }
  }
  //if (progress)
  //  fprintf("\nOutput written to output file\n\n");
}
void pdis_class::startup(void) {

  char outfilename[100];  //infilename[100],

  if ( printdata || printdis ) {
    openfile(&outfile,PDIS_OUTFILE,"w",outfilename);
    if ( printdata ) {
      printHeadings();
    }
  }

  makedists();

  if ( printdata || printdis ) {
    FClose(outfile);
  }
}




