/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#include "tree_class.h"

tree_class::tree_class() {}
tree_class::~tree_class() {}
void tree_class::setupTree( Int4 sp, boolean njoining )
{
  njoin = njoining;
  numsp = sp;
  numsp1 = numsp + 1;
  numsp2 = numsp * 2 - 2;

  /* initialize a tree */
  Int4 i, j;
  node_class *p, *q;
  nodep = new node_class * [numsp2+1];
  for (i = 0; i < numsp; i++) {
    nodep[i] = new node_class;
    nodep[i]->tip = true;
    nodep[i]->number = i + 1;
    nodep[i]->v = 0.0;
  }
  for (i = numsp1; i <= numsp2; i++) {
    q = NULL;
    for (j = 1; j <= 3; j++) {
      p = new node_class;
      p->tip = false; p->number = i; p->next = q; q = p;
    }
    p->next->next->next = p;
    nodep[i - 1] = p;
  }
  q = NULL;
  for (j = 1; j <= 2; j++) {
    p = new node_class;
    p->tip = false; p->number = numsp2 + 1; p->next = q; q = p;
  }
  p->next->next = p; nodep[numsp2] = p; start = nodep[0];
}
int tree_class::printTree( char *filename )
{

  openfile ( &outfile, filename, "w" );
  if ( outfile == NULL ) return 1;

  outgropt = false;  /* One option that must be set to print the tree */
  printTreeDiag();
  FClose(outfile);
#ifdef DEBUG
  printf("printTree: %s has been created.\n", filename);
#endif
  return 0;

}
int tree_class::printTree( FILE * file ) 
{
  if ( file == NULL ) { return 1; }

  outgropt = false;  /* One option that must be set to print the tree */
  outfile = file;
  printTreeDiag();
#ifdef DEBUG
  printf("printTree: The file has been created.\n");
#endif
  return 0;
}
int tree_class::printTreeFile( char *filename )
{
  openfile ( &treefile, filename, "w" );
  if ( treefile == NULL ) { return 1; }

  col = 0;
  treeout(start);
  FClose(treefile);
#ifdef DEBUG
  printf("printTreeFile: %s has been created.\n", filename);
#endif
  return 0;
}
int tree_class::printTreeFile( FILE *file )
{
  if ( file == NULL ) { return 1; }

  col = 0;
  treefile = file;
  treeout(start);
#ifdef DEBUG
  printf("printTreeFile: The file has been created.\n");
#endif
  return 0;
}
bTreeNode *tree_class::getBTree()
{
  bTreeNode *bnode = 0;
  bTreeOut(start, &bnode);
  return(bnode);
}
void tree_class::openfile(FILE **fp, char *filename, const char *mode )
{
    (*fp) = fopen(filename,mode); /**/
}
void tree_class::coordinates(node_class *p,
			     double lengthsum,
			     Int4 *tipy,
			     double *tipmax)
{  /* establishes coordinates of nodes */
  node_class *q, *first, *last;

  if (p->tip) {
    p->xcoord = (Int4)(over * lengthsum + 0.5);
    p->ycoord = (*tipy);
    p->ymin = (*tipy);
    p->ymax = (*tipy);
    (*tipy) += down;
    if (lengthsum > (*tipmax))
      (*tipmax) = lengthsum;
    return;
  }
  q = p->next;
  do {
    coordinates(q->back, lengthsum + q->v, tipy,tipmax);
    q = q->next;
  } while ((p == start || p != q)
        && (p != start || p->next != q));
  first = p->next->back; q = p;
  while (q->next != p) q = q->next;
  last = q->back;
  p->xcoord = (Int4)(over * lengthsum + 0.5);
  if (p == start) {
    if (njoin) p->ycoord = p->next->next->back->ycoord;
    else p->ycoord = (p->back->ycoord + p->next->back->ycoord) / 2;
  } else p->ycoord = (first->ycoord + last->ycoord) / 2;
  p->ymin = first->ymin;
  p->ymax = last->ymax;
}
void tree_class::drawline(Int4 i, double scale)
{
  /* draws one row of the tree diagram by moving up tree */
  node_class *p, *q;
  Int4 n, j;
  boolean extra, trif;
  node_class *r, *first, *last;
  boolean done;

  p = start;
  q = start;
  extra = false;
  trif = false;
  if (i == p->ycoord && p == start) {
    if (p->number - numsp >= 10)
      fprintf(outfile, "-%2ld", /*p->number - numsp*/p->number);
    else fprintf(outfile, "--%d", /*p->number - numsp*/p->number);
    extra = true; trif = true;
  } else fprintf(outfile, "  ");
  do {
    if (!p->tip) {
      r = p->next; done = false;
      do {
        if (i >= r->back->ymin && i <= r->back->ymax) {
          q = r->back; done = true;
        }
        r = r->next;
      } while (!(done || (p != start && r == p) ||
                 (p == start && r == p->next)));
      first = p->next->back;
      r = p;
      while (r->next != p) r = r->next;
      last = r->back;
      if (p == start) last = p->back;
    }
    done = (p->tip || p == q);
    n = (Int4)(scale * (q->xcoord - p->xcoord) + 0.5);
    if (n < 3 && !q->tip) n = 3;
    if (extra) { n--; extra = false; }
    if (q->ycoord == i && !done) {
      if (p->ycoord != q->ycoord) putc('+', outfile);
      if (trif) { n++; trif = false; }
      if (!q->tip) {
        for (j = 1; j <= n - 2; j++) putc('-', outfile);
        if (q->number - numsp >= 10)
          fprintf(outfile, "%2ld",q->number);
        else fprintf(outfile, "-%d",q->number);
        extra = true;
      } else {
        for (j = 1; j < n; j++) putc('-', outfile);
      }
    } else if (!p->tip) {
      if (last->ycoord > i && first->ycoord < i && i != p->ycoord) {
        putc('!', outfile);
        for (j = 1; j < n; j++) putc(' ', outfile);
      } else {
        for (j = 1; j <= n; j++) putc(' ', outfile);
        trif = false;
      }
    }
    if (q != p) p = q;
  } while (!done);
  if (p->ycoord == i && p->tip) {
    fprintf(outfile, "%-10ld", p->number);
  }
  putc('\n', outfile);
}
void tree_class::describe( node_class *p)
{
  /* print out information for one branch */
  Int4 i;
  node_class *q;

  q = p->back;
  fprintf(outfile, "%4ld          ", q->number);
  if (p->tip) {
    fprintf(outfile, "%4ld      ",p->number);
  } else {
    fprintf(outfile, "%4ld      ", p->number);
  }
  fprintf(outfile, "%15.5f\n", q->v);
  if (!p->tip) { describe(p->next->back); describe(p->next->next->back); }
}
void tree_class::summarize(void) 
{
  /* print out branch lengths etc. */
  putc('\n', outfile);
  if (njoin) {
    fprintf(outfile, "remember:");
    if (outgropt) fprintf(outfile, " (although rooted by outgroup)");
    fprintf(outfile, " this is an unrooted tree!\n");
  }
  fprintf(outfile, "\nBetween        And            Length\n");
  fprintf(outfile, "-------        ---            ------\n");
  describe(start->next->back);
  describe(start->next->next->back);
  if (njoin) describe(start->back);
  fprintf(outfile, "\n\n");
}
void tree_class::treeout( node_class *p )
{
  /* write out file with representation of final tree */
  Int4 i, n, w;
  char c;
  double x;

  if (p->tip) {
    n = 0;
    fprintf(treefile, "%d", p->number);
    col += n;
  } else {
    putc('(', treefile); col++; treeout(p->next->back);
    putc(',', treefile); col++;
    if (col > 55) { putc('\n', treefile); col = 0; }
    treeout(p->next->next->back);
    if (p == start && njoin) { putc(',', treefile); treeout(p->back); }
    putc(')', treefile); col++;
  }
  x = p->v;
  if (x > 0.0) w = (Int4)(0.43429448222 * log(x));
  else if (x == 0.0) w = 0;
  else w = (Int4)(0.43429448222 * log(-x)) + 1;
  if (w < 0) w = 0;
  if (p == start) fprintf(treefile, ";\n");
  else {
    fprintf(treefile, ":%*.5f", (int)(w + 7), x); col += w + 8;
  }
}
double tree_class::bTreeOut( node_class *p, bTreeNode **node)
{ /* create bTree with representation of final tree */
  double x;

  if ( !(p->tip) ) {
      bTreeNode *leftPtr, *rightPtr;
      double left, right;

      left = bTreeOut(p->next->back, &leftPtr );
      right = bTreeOut(p->next->next->back, &rightPtr ); 
      if (p == start && njoin) {
	double     leftDisTemp;
	bTreeNode *leftPtrTemp;
	bTreeNode *nodetemp = new bTreeNode(leftPtr,rightPtr,left,right,p->number);
	leftDisTemp = bTreeOut(p->back, &leftPtrTemp );
	(*node) = new bTreeNode(leftPtrTemp, nodetemp, leftDisTemp, 0, p->number+1); 
      } else  {
	(*node) = new bTreeNode(leftPtr, rightPtr, left, right, p->number);
      }
  } else {
    (*node) = new bTreeNode(p->number);
  }
  x = p->v;
  if (p == start) {
#ifdef DEBUG
    printf("bTreeOut: The bTree has been created.\n");
#endif
    return (0);
  } else {
    return x;
  }
}
void tree_class::printTreeDiag() 
{
  /* prints out diagram of the tree */
  Int4 tipy = 1;
  double scale, tipmax = 0.0;
  Int4 i;

  putc('\n', outfile);
  coordinates( start, 0.0, &tipy,&tipmax);
  scale = 1.0 / (Int4)(tipmax + 1.000);
  for (i = 1; i <= tipy - down; i++) drawline(i, scale);
  putc('\n', outfile);
  summarize();
}
