/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/* tree_class.c 08/17/99 
   Modified by Trevor Carlson (tcarlson@andrew.cmu.edu)
   The Original program can be found at:
     http://evolution.genetics.washington.edu/phylip.html

   This class is directly used by the nbr_class.c file.  The neighbor method
 of creating trees is used when this tree is created.  This program is between
 the nbr_class.c and bTreeNode.c file.  This class was originally created to take
 advantage of his tree structure.  Aftwerwards, a new tree class <bTreeNode.c>
 was created to allow other functions of a tree to me handled.

 ***********WARNING***********
  There is no proper destructor for this class, nor is there a proper copy constructor.
  Garbage will be produced after it leaves scope.
 ***********WARNING***********

  bTreeNode *getBTree() - this accessor creates a new bTreeNode that contains the exact 
   same tree structure that is contained inside this tree_class.  

*/
/* version 3.52c. (c) Copyright 1993 by Joseph Felsenstein and Mary Kuhner.
   Written by Mary Kuhner, Jon Yamato, Joseph Felsenstein, Akiko Fuseki,
   Sean Lamont, and Andrew Keeffe.
   Permission is granted to copy and use this program provided no fee is
   charged for it and provided that this copyright notice is not removed. */
#ifndef __tree_class_h__
#define __tree_class_h__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "phylip.h"
#include "node_class.h"
#include "bTreeNode.h"

class tree_class {

 public: 

           tree_class();
	  ~tree_class();
      void setupTree( Int4, boolean );

      int  printTree( char * );
      int  printTree( FILE * );
      int  printTreeFile( char * );
      int  printTreeFile( FILE * );

bTreeNode *getBTree();

  node_class **nodep;
  node_class *start;

 private:

  boolean njoin, outgropt;
  Int4 numsp, numsp1, numsp2, col;
  FILE *outfile, *treefile;

  void openfile(FILE **, char *, const char * );
  void coordinates(node_class *,
	            double,
		    Int4 *,
		    double *);
  void drawline(Int4 , double);
  void describe( node_class * );
  void summarize ( void );
  void treeout( node_class * );
double bTreeOut( node_class *, bTreeNode ** );
  void printTreeDiag();

};

#endif 

